# Author: Lars Brenna <larsb@stud.cs.uit.no>
# Date started: Thu 17 Jun 2004

import threading, traceback, SOAPpy, logging
    

class Handler(threading.Thread):
    """WAIFService.Handler defines the minimum recommended
    Handler API for a WAIF Service supporting PONS configuration, as
    developed by Lars Brenna, 2004.  A WAIF Service handler is a thread
    handling incoming events by blocking on a synchronized queue.

    Users of this class must add custom handlers to implement the actual
    functionality of the WAIF service.

    Handler.start() must be called from parent after registering handlers
    """

    def __init__(self, parent, name, queue, logger=None):
        """Handler.__init__ returns a Handler object.
        Needs two arguments; the name of the WAIF service in case a logger is
        not defined, and a queue to receive incoming events from.
        If no logger is provided, a standard file log facility is created.

        """
        threading.Thread.__init__(self)
        self.setDaemon(1)

        self.parent = parent
        self.eventQ = queue

        if not logger:
            # setting up log facility
            self.logger = logging.getLogger(name)
            hdlr = logging.FileHandler('%s.log' % name)
            formatter = logging.Formatter('%(asctime)s %(levelname)s %(message)s')
            hdlr.setFormatter(formatter)
            self.logger.addHandler(hdlr) 
        else:
            self.logger = logger
            
        self.name = name

        
        self.handlerHash = {}

        print "%s.Handler init ok!" % name
        self.logger.info("%s.Handler init ok!" % name)

    def registerFunction(self, function, *handles):
        """
        Users of the WAIFService.Handler class can register custom event
        handlers via this function.

        A single handler/function could be assigned to several handles, e.g.
        the same function can implement the handling of a range of datatypes.
        This also loosens the semantics of the system, allowing a less strict
        typing of event contents, or in the other end; event handling based on
        MIME-types or similar.

        The 'noAction' handler is the default handler if no other handlers are
        registered.
        """
        if not len(handles):
            self.logger.warning("Registering event handler on its own name: %s"\
                                % function.__name__ )
            self.handlerHash[function.__name__] = function
        else:
            for handle in handles:
                self.handlerHash[handle] = function
            self.logger.info("Registered %d handles for %s: %s"\
                             % (len(handles), function.__name__, handles) )
        
    

    def run(self):
        """Main runtime of the WAIF Service Handler.
        Blocks on a Queue for incoming events. If a handler is associated
        to the event format, it is found via the handlerHash table.
        """
        self.logger.debug("Handler entering getEvent-loop")
        while(1):
            try:
                event = self.eventQ.get()
                self.logger.debug("%s.Handler got event!" % self.name )
                self.handlerHash.get( event[2], self.noAction )(event)
                
            except:
                self.logger.exception("EventHandler FAILED.")
                self.logger.info("Attempting to use the default event handler")
                self.noAction(event)




    def noAction(self, event):
        """The default event handler when no other are registered or
        suitable. """
        #if self.handlerHash.has_key(self.noAction.__name__):
        #    selv.handlerHash[self.noAction.__name__]( event )        
        #else:    
        self.logger.warning( "HANDLER: No action for event: %s"\
                             % str(event))
        return
        

    def push(self, taddr, waifID, targetID, data=None):
        """Push msg to taddr"""
        host, port = taddr
        target = SOAPpy.SOAPProxy("http://%s:%d/" % (host, port))

        if data:
            target.dispatch(waifID, targetID, data)
        else:
            target.dispatch(waifID, targetID)
            
        self.logger.debug("Pushed event %s to %s" %(data, taddr))

