# Author: Lars Brenna <larsb@stud.cs.uit.no>
# Date started: Thu 17 Jun 2004

import SOAPpy, logging, threading, Queue
from random import randint
from socket import gethostname, gethostbyaddr


class Listener:
    """WAIFService.Listener defines the minimum API required for a WAIF
    Service listener supporting PONS configuration, as developed by
    Lars Brenna, 2004.
    A WAIF Service Listener is a thread exporting the WAIF Service external
    API through a SOAP server instance. It accepts incoming events, and queue
    them to a WAIF Service Handler instance. Custom handlers can be dynamically
    added to the Handler thread. Creates a log called ('%s.log' % name),
    referred to as self.logger
    """

    def __init__(self, name, port=None, host=None, logger=None):
        if not port:
            port = 8050
        if not host:
            self.mehost = (gethostbyaddr(gethostname())[0], port)
        else:
            self.mehost = (host, port)
        self.users = {}
        self.name = name

        # Setting the default event handler
        self.defaultHandler = 'noAction'

        if not logger:
            # setting up log facility
            self.logger = logging.getLogger(name)
            hdlr = logging.FileHandler('%s.log' % name)
            formatter = logging.Formatter('%(asctime)s %(levelname)s %(message)s')
            hdlr.setFormatter(formatter)
            self.logger.addHandler(hdlr)
            self.logger.setLevel(logging.DEBUG)
        else:
            self.logger = logger

        try:
            self.server = SOAPpy.SOAPServer(self.mehost)
        except:
            self.logger.exception("Could not create SOAP Server instance..")
            self.logger.error("Unable to start WAIF Service %s" % name )
            self.logger.error("Exiting..")
            raise Exception, "Unable to start WAIF Service %s" % name
        

        self.jobQueue = Queue.Queue()
        self.logger.info( "WAIF %s service initialized at %s"
                          % (name, self.mehost) )

        print "%s.Listener init OK on host %s, logging to: %s.log" % \
              (name, self.mehost, name)
        
    

    
    def run(self):
        """This method starts the SOAP server and exposes the API to
        the address in self.mehost. Additional methods could be exposed by
        manually registering them to server.registerFunction or .registerObject
        before calling server.serve_forever(), which is an infinite looping
        function.
        """
               
        
        # exposing the methods of this service:
        self.server.registerFunction(self.subscribe)
        self.server.registerFunction(self.unsubscribe)
        self.server.registerFunction(self.dispatch)
    

        try:
            self.logger.debug("Entering Listener.serve_forever() loop..")
            self.server.serve_forever()
        
        finally:
            self.logger.info( "exiting..")

        self.logger.shutdown()



    def subscribe(self, waifID, taddr, params={} ):
        """Register user if new, add and configure new subscription.
        A profile is created in self.users[waifID][subID]['profile'] by the
        customizable self.profiler(..) function.
        Additional information is also stored in the dictionary found by
        [waifID][subID]. This info can be read and written openly.
        """
        # converting from SOAP associative array to python dict
        try:
            preferences = params._asdict()
        except:
            self.logger.exception("Received weird params: [%s , %s]" % ( params, type(params)))
            self.logger.info("Using default subscription configuration")
            preferences = {}

        # if waifID not present, create new entry
        self.users[waifID] = self.users.get(waifID, {} )
        # creating new subID
        subID = randint(0,1024)
        
        self.users[waifID][subID] = self.profiler(waifID, subID, taddr, preferences)
              
        self.logger.info( "Installed subscription from %s on %s with ID %d"
                          % (waifID, list(taddr), subID))
        return subID


    def profiler(self, waifID, subID, taddr, sub_params):
        """Creating the default profile.
        This function should be overridden to change how user profiles are
        created. Programmers should use this implementation as a starting point
        """
        profile = {}
        profile['taddr'] = taddr
        profile['remoteID'] = sub_params.pop('remoteID', 0)
        if not profile['remoteID']:
            self.logger.warning("PROFILER: parameter remoteID did not follow subscription, using 0")

        # setting the default event handler for this subscription
        profile['handle']  = sub_params.pop('DataType', self.defaultHandler )

        # if there are extra data in the subscription parameters that are needed during
        # event handling, these should be stored as 'preferences' or similar variable.
        # otherwise, preferences == {}
        profile['preferences'] = sub_params

        return profile


    def unsubscribe(self, waifID, subID):
        """removes subcription. If it is the owners only subscription
        the owner is removed from our lists too. 
        """
        try:
            del self.users[waifID][subID]
            answer = "user %s cancelled subscription %d to WAIF %s service"\
                     % (waifID, subID, self.name)
            self.logger.info( answer )
        except KeyError:
            answer = "In unsubscribe: [%s:%d] No such waifID:subID combination"\
                     % (waifID, subID)
            self.logger.exception( answer )
        
        return answer
    
        
    # trigger event push to user waifID
    def dispatch(self, waifID, subID, message=None):
        """dispatch is called to receive an event. Needs a WAIFID and an
        alertID to identify a subscription. The optional message argument
        can contain data to be handled by the WAIF service, and should as such
        be formatted as a Python dictionary or compatible data structure.
        Returns 1 when message is successfully received and queued, otherwise
        -1 or an exception.
        """

        if not self.users[waifID].has_key(subID):
            self.logger.error( "[%s:%d] No such waifID:subID combination"
                               % (waifID, subID) )
            return -1

        if message: # it should be a dictionary
            try:
                message = message.__asdict()
                handle = message.get('DataType', self.users[waifID][subID]['handle'])
            except:
                self.logger.exception("Weird event format, trying anyway.")
                
        else:
            handle = self.users[waifID][subID]['handle']
            
        task = (waifID, subID, handle, message)
        self.jobQueue.put(task)
        return 1
    

